/*  Copyright (c) MIPoPS. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-3-Clause license that can
 *  be found in the LICENSE file in the root of the source tree.
 */

//---------------------------------------------------------------------------
#include "Common/ProcessFile.h"
#include "ZenLib/Ztring.h"
#include "Output.h"
#include <iostream>
#include <mutex>
#include <iomanip>
#include <map>
#include "TimeCode.h"
using namespace ZenLib;
//---------------------------------------------------------------------------

//***************************************************************************
// Callback
//***************************************************************************

void __stdcall Event_CallBackFunction(unsigned char* Data_Content, size_t Data_Size, void* UserHandler_Void)
{
    //Retrieving UserHandler
    file*                               UserHandler = (file*)UserHandler_Void;
    struct MediaInfo_Event_Generic*     Event_Generic = (struct MediaInfo_Event_Generic*) Data_Content;
    unsigned char                       ParserID;
    unsigned short                      EventID;
    unsigned char                       EventVersion;

    //Integrity test
    if (Data_Size < 4)
        return; //There is a problem

    //Retrieving EventID
    ParserID = (unsigned char)((Event_Generic->EventCode & 0xFF000000) >> 24);
    EventID = (unsigned short)((Event_Generic->EventCode & 0x00FFFF00) >> 8);
    EventVersion = (unsigned char)(Event_Generic->EventCode & 0x000000FF);
    switch (ParserID)
    {
    case MediaInfo_Parser_DvDif:
        switch (EventID)
        {
        case MediaInfo_Event_DvDif_Analysis_Frame: if (EventVersion == 1 && Data_Size >= sizeof(struct MediaInfo_Event_DvDif_Analysis_Frame_1)) UserHandler->AddFrame((MediaInfo_Event_DvDif_Analysis_Frame_1*)Event_Generic); break;
        case MediaInfo_Event_DvDif_Change: if (EventVersion == 0 && Data_Size >= sizeof(struct MediaInfo_Event_DvDif_Change_0)) UserHandler->AddChange((MediaInfo_Event_DvDif_Change_0*)Event_Generic); break;
        case MediaInfo_Event_Global_Demux: if (EventVersion == 4 && Data_Size >= sizeof(struct MediaInfo_Event_Global_Demux_4)) UserHandler->AddFrame((MediaInfo_Event_Global_Demux_4*)Event_Generic); break;
        }
        break;
    }
}

//***************************************************************************
// Constructor/Destructor
//***************************************************************************

static int Pos = 0;
//---------------------------------------------------------------------------
file::file()
{
    FilePos = Pos++;

    FrameNumber = 0;
}

//---------------------------------------------------------------------------
void file::Parse(const String& FileName)
{
    MI.Option(__T("File_Event_CallBackFunction"), __T("CallBack=memory://") + Ztring::ToZtring((size_t)&Event_CallBackFunction) + __T(";UserHandler=memory://") + Ztring::ToZtring((size_t)this));
    MI.Option(__T("File_DvDif_Analysis"), __T("1"));
    MI.Open(FileName);

    // Filing some info
    FrameRate = Ztring(MI.Get(Stream_Video, 0, __T("FrameRate_Original"))).To_float64();
    if (!FrameRate)
        FrameRate = Ztring(MI.Get(Stream_Video, 0, __T("FrameRate"))).To_float64();
    if (!FrameRate || (FrameRate >= 29.97 && FrameRate <= 29.98))
        FrameRate = double(30 / 1.001); // Default if no frame rate available, or better rounding
}

//---------------------------------------------------------------------------
file::~file()
{
    for (auto& Frame : PerFrame)
    {
        delete[] Frame->Errors;
        delete[] Frame->Video_STA_Errors;
        delete[] Frame->Audio_Data_Errors;
        delete Frame;
    }
    for (auto& Change : PerChange)
    {
        delete Change;
    }
}

//***************************************************************************
// Helpers
//***************************************************************************

//---------------------------------------------------------------------------
void file::AddChange(const MediaInfo_Event_DvDif_Change_0* FrameData)
{
    FrameNumber++; // Event FrameCount is currently wrong

    // Check if there is a change we support
    if (!PerChange.empty())
    {
        const auto Current = PerChange.back();
        if (Current->Width == FrameData->Width
            && Current->Height == FrameData->Height
            && Current->VideoChromaSubsampling == FrameData->VideoChromaSubsampling
            && Current->VideoScanType == FrameData->VideoScanType
            && Current->VideoRatio_N == FrameData->VideoRatio_N
            && Current->VideoRatio_D == FrameData->VideoRatio_D
            && Current->VideoRate_N == FrameData->VideoRate_N
            && Current->VideoRate_D == FrameData->VideoRate_D
            && Current->AudioRate_N == FrameData->AudioRate_N
            && Current->AudioRate_D == FrameData->AudioRate_D
            && Current->AudioChannels == FrameData->AudioChannels
            && Current->AudioBitDepth == FrameData->AudioBitDepth
            && Current->Captions_Flags == FrameData->Captions_Flags)
        {
            return;
        }
    }

    MediaInfo_Event_DvDif_Change_0* ToPush = new MediaInfo_Event_DvDif_Change_0();
    std::memcpy(ToPush, FrameData, sizeof(MediaInfo_Event_DvDif_Change_0));
    ToPush->FrameNumber = FrameNumber - 1; // Event FrameCount is currently wrong
    PerChange.push_back(ToPush);
}

vector<string> FileNames;
string Target;
mutex Mutex;
enum status : uint8_t
{
    Status_OK,
    Status_Problem,
    Status_TimeCodeIssue, // e.g. time code missing
    Status_Skipped, // e.g. time code jump
};
vector<TimeCode> Frames_TimeCode[100];
vector<status> Frames_Status[100];
static size_t Frames_Status_Pos = 0;
vector<size_t> Count_Frames_Healthy;
vector<size_t> Count_Frames_NotHealthy_Blocks_NotHealthy;
vector<size_t> Count_Frames_Used;
vector<size_t> Count_Blocks_Healthy;
vector<size_t> Count_Blocks_NotHealthy;
vector<size_t> Count_Blocks_Used;
vector<size_t> DoNotUseFile;
size_t DoNotUseFiles = 0;
vector<MediaInfo_int8u*> Mergeable[100];
vector<size_t> Mergeable_Counts;
size_t Frames_Good = 0;
size_t Frames_Recovered = 0;
size_t Blocks_NotFixed = 0;
FILE* F[100+1];
int FTell[100+1];
uint8_t Buffers[100+1][120000];
size_t Frames_Status_Max = -1;

const int FrameCount_Width = 4;
const int BlockCount_Width = 8;
static int Precision = 6;
const bool Verbose = true;

//---------------------------------------------------------------------------
void file::AddFrame(const MediaInfo_Event_DvDif_Analysis_Frame_1* FrameData)
{
    MediaInfo_Event_DvDif_Analysis_Frame_1* ToPush = new MediaInfo_Event_DvDif_Analysis_Frame_1();
    std::memcpy(ToPush, FrameData, sizeof(MediaInfo_Event_DvDif_Analysis_Frame_1));
    if (FrameData->Errors)
    {
        size_t SizeToCopy = std::strlen(FrameData->Errors) + 1;
        auto Errors = new char[SizeToCopy];
        std::memcpy(Errors, FrameData->Errors, SizeToCopy);
        ToPush->Errors = Errors;
    }
    if (FrameData->Video_STA_Errors)
    {
        size_t SizeToCopy = FrameData->Video_STA_Errors_Count * sizeof(size_t);
        auto Video_STA_Errors = new size_t[SizeToCopy];
        std::memcpy(Video_STA_Errors, FrameData->Video_STA_Errors, SizeToCopy);
        ToPush->Video_STA_Errors = Video_STA_Errors;
    }
    if (FrameData->Audio_Data_Errors)
    {
        size_t SizeToCopy = FrameData->Audio_Data_Errors_Count * sizeof(size_t);
        auto Audio_Data_Errors = new size_t[SizeToCopy];
        std::memcpy(Audio_Data_Errors, FrameData->Audio_Data_Errors, SizeToCopy);
        ToPush->Audio_Data_Errors = Audio_Data_Errors;
    }
    PerFrame.push_back(ToPush);

    if (Target.empty())
        return;
        
    Mutex.lock();
    auto FileCount = FileNames.size();
    if (DoNotUseFile.empty())
    {
        DoNotUseFile.resize(FileCount);
        Count_Frames_Healthy.resize(FileCount);
        Count_Frames_NotHealthy_Blocks_NotHealthy.resize(FileCount);
        Count_Frames_Used.resize(FileCount);
        Count_Blocks_Healthy.resize(FileCount);
        Count_Blocks_NotHealthy.resize(FileCount);
        Count_Blocks_Used.resize(FileCount);
        Mergeable_Counts.resize(FileCount);

        for (size_t i = 0; i < FileCount; i++)
            F[i] = fopen(FileNames[i].c_str(), "rb");
        F[FileCount] = fopen(Target.c_str(), "wb");
    }

    // Stop if frame repetition was found
    if (DoNotUseFile[FilePos] || timecode(FrameData->TimeCode).Repeat())
    {
        if (!DoNotUseFile[FilePos])
        {
            DoNotUseFile[FilePos] = true;
            DoNotUseFiles++;
        }
        Mutex.unlock();
        return;
    }

    // Time code jumps - after first frame
    timecode TC_Temp(FrameData->TimeCode);
    TimeCode TC_Current;
    if (TC_Temp.HasValue())
        TC_Current = TimeCode(TC_Temp.TimeInSeconds() / 3600, (TC_Temp.TimeInSeconds() / 60) % 60, TC_Temp.TimeInSeconds() % 60, TC_Temp.Frames(), 30 /*TEMP*/, TC_Temp.DropFrame());
    status StatusIfNoProblem = Status_OK;
    if (!Frames_TimeCode[FilePos].empty())
    {
        TimeCode TC_Previous(Frames_TimeCode[FilePos].back());
        TC_Previous++;
        if (TC_Current.HasValue())
        {
            while (TC_Previous != TC_Current)
            {
                Frames_TimeCode[FilePos].push_back(TC_Previous);
                Frames_Status[FilePos].push_back(Status_Skipped);
                Mergeable[FilePos].push_back(nullptr);
                TC_Previous++;
            }
        }
        else
        {
            TC_Current = TC_Previous;
            StatusIfNoProblem = Status_TimeCodeIssue;
        }
    }
    Frames_TimeCode[FilePos].push_back(TC_Current);
    
    // Stats
    if (!(FrameData->Video_STA_Errors || FrameData->Audio_Data_Errors))
        Count_Frames_Healthy[FilePos]++;
    for (int i = 0; i < 1500; i++)
    {
        switch (FrameData->Mergeable[i])
        {
            case 1: Count_Blocks_Healthy[FilePos]++; break;
            case 2: Count_Blocks_NotHealthy[FilePos]++; break;
            default:;
        }
    }
    Frames_Status[FilePos].push_back((FrameData->Video_STA_Errors || FrameData->Audio_Data_Errors) ? Status_Problem : StatusIfNoProblem);

    // Mergeable blocks
    auto Mergeable_Item = new MediaInfo_int8u[1500];
    memcpy(Mergeable_Item, FrameData->Mergeable, 1500);
    Mergeable[FilePos].push_back(Mergeable_Item);

    // Time code jumps - first frame
    if (!Frames_Status_Pos)
    {
        Frames_Status_Max = -1;
        for (size_t i = 0; i < FileCount; i++)
            if (!DoNotUseFile[FilePos] && Frames_Status_Max > Frames_Status[i].size())
                Frames_Status_Max = Frames_Status[i].size();
        if (Frames_Status_Max)
        {
            TimeCode TC_Min = Frames_TimeCode[0][0];
            TimeCode TC_Max = TC_Min;
            for (size_t i = 0; i < FileCount; i++)
            {
                if (TC_Min.ToFrames() > Frames_TimeCode[i][0].ToFrames())
                    TC_Min = Frames_TimeCode[i][0];
                if (TC_Max.ToFrames() < Frames_TimeCode[i][0].ToFrames())
                    TC_Max = Frames_TimeCode[i][0];
            }
            if (TC_Min != TC_Max)
                for (size_t i = 0; i < FileCount; i++)
                {
                    auto TC_Previous = TC_Min;
                    auto TC_End = Frames_TimeCode[i].front();
                    while (TC_Previous != TC_End)
                    {
                        Frames_TimeCode[i].insert(Frames_TimeCode[i].begin(), TC_Previous);
                        Frames_Status[i].insert(Frames_Status[i].begin(), Status_Skipped);
                        Mergeable[i].insert(Mergeable[i].begin(), nullptr);
                        TC_Previous++;
                    }
                }
        }
    }

    // Check which frames are available everywhere
    Frames_Status_Max = -1;
    for (size_t i = 0; i < FileCount; i++)
        if (!DoNotUseFile[FilePos] && Frames_Status_Max > Frames_Status[i].size())
            Frames_Status_Max = Frames_Status[i].size();
    DisplayLines();
    Mutex.unlock();
}

//---------------------------------------------------------------------------
void file::DisplayLines()
{
    auto FileCount = FileNames.size();

    for (; Frames_Status_Pos < Frames_Status_Max; Frames_Status_Pos++)
    {
        if (Verbose)
            cout << setfill(' ') << setw(FrameCount_Width) << Frames_Status_Pos << " ";

        // Copy 1 frame in memory, for each input
        for (size_t i = 0; i < FileCount; i++)
        {
            if (Frames_Status_Pos < Frames_Status[i].size())
            {
                auto B = fread(Buffers[i], 1, 120000, F[i]);
                if (B != 120000)
                    int A = 0;
            }
        }

        // Find a valid frame
        size_t Prefered = -1;
        for (size_t i = 0; i < FileCount; i++)
        {
            if (Frames_Status_Pos < Frames_Status[i].size() && Frames_Status[i][Frames_Status_Pos] == Status_OK)
            {
                Prefered = i;
                break;
            }
        }
        if (Prefered == -1)
        {
            for (size_t i = 0; i < FileCount; i++)
            {
                if (Frames_Status_Pos < Frames_Status[i].size() && Frames_Status[i][Frames_Status_Pos] == Status_TimeCodeIssue)
                {
                    Prefered = i;
                    break;
                }
            }
        }
        memcpy(Buffers[FileCount], Buffers[0], 120000);
        if (Prefered != -1)
        {
            if (Verbose)
                cout << Prefered;
            Count_Frames_Used[Prefered]++;
            Frames_Good++;
            if (Prefered)
            {
                memcpy(Buffers[FileCount], Buffers[Prefered], 120000);
                Count_Blocks_Used[Prefered] += 1500;
            }
            else
                Count_Blocks_Used[0] += 1500;
        }
        else
        {
            if (Verbose)
                cout << "X";
        }
        if (Verbose)
        {
            cout << "   (";
            for (size_t i = 0; i < FileCount; i++)
                if (Frames_Status_Pos >= Frames_Status[i].size())
                    cout << 'N';
                else if (Frames_Status[i][Frames_Status_Pos] == Status_Problem)
                    cout << 'P';
                else if (Frames_Status[i][Frames_Status_Pos] == Status_Skipped)
                    cout << 'S';
                else if (Frames_Status[i][Frames_Status_Pos] == Status_TimeCodeIssue)
                    cout << 'T';
                else
                    cout << ' ';
            cout << ')';

            for (size_t i = 0; i < FileCount; i++)
            {
                string TC;
                if (Frames_Status_Pos < Frames_Status[i].size())
                {
                    if (Frames_Status[i][Frames_Status_Pos] == Status_TimeCodeIssue)
                        TC = "XX:XX";
                    else
                        TC = Frames_TimeCode[i][Frames_Status_Pos].ToString();
                }
                if (TC.empty())
                    TC = "     ";
                else if (TC != "XX:XX")
                    TC = TC.substr(6);

                cout << ' ' << TC;
            }
        }

        // Find valid blocks
        if (Prefered == -1)
        {
            Count_Frames_Used[0]++; // By default

            cout << ", ";
            size_t ThisFrame_Count_Blocks_Healthy[100];
            size_t ThisFrame_Count_Blocks_NotHealthy[100];
            size_t ThisFrame_Count_Blocks_Used[100];
            memset(ThisFrame_Count_Blocks_Healthy, 0, FileCount * sizeof(size_t));
            memset(ThisFrame_Count_Blocks_NotHealthy, 0, FileCount * sizeof(size_t));
            memset(ThisFrame_Count_Blocks_Used, 0, FileCount * sizeof(size_t));
            size_t IssueCount = 0;

            // For each block
            vector<size_t> Priorities; // Find the frame with the minimum count of bad blocks
            Priorities.resize(FileCount);
            for (int i = 0; i < 1500; i++)
            {
                for (int j = 0; j < FileCount; j++)
                    if (Frames_Status_Pos < Frames_Status[j].size() && Frames_Status[j][Frames_Status_Pos])
                    {
                        switch (Mergeable[j][Frames_Status_Pos][i])
                        {
                        case 2:
                            Priorities[j]++;
                            break;
                        default:;
                        }
                    }
                    else
                        Priorities[j] = 1500;
            }
            map<size_t, vector<size_t>> Priorities2;
            for (int k = 0; k < FileCount; k++)
                Priorities2[Priorities[k]].push_back(k);
            Priorities.clear();
            for (auto p : Priorities2)
                for (auto k : p.second)
                    Priorities.push_back(k);
            ThisFrame_Count_Blocks_Used[Priorities[0]] = 1500;
            if (Priorities[0])
                memcpy(Buffers[FileCount], Buffers[Priorities[0]], 120000); // Copy the content of the file having the less issues
            for (int i = 0; i < 1500; i++)
            {
                map <size_t, size_t> B;
                bool NoIssue = false;
                for (int k = 0; k < FileCount; k++)
                {
                    auto j = Priorities[k];
                    if (Frames_Status_Pos < Frames_Status[j].size() && Frames_Status[j][Frames_Status_Pos])
                    {
                        B[Mergeable[j][Frames_Status_Pos][i]]++;
                        switch (Mergeable[j][Frames_Status_Pos][i])
                        {
                        case 1:
                            ThisFrame_Count_Blocks_Healthy[j]++;
                            if (!NoIssue)
                            {
                                NoIssue = true;
                                if (k)
                                {
                                    memcpy(Buffers[FileCount] + i * 80, Buffers[j] + i * 80, 80);
                                    ThisFrame_Count_Blocks_Used[Priorities[0]]--;
                                    ThisFrame_Count_Blocks_Used[j]++;
                                }
                            }
                            break;
                        case 2:
                            ThisFrame_Count_Blocks_NotHealthy[j]++;
                            break;
                        default:;
                            NoIssue = true;
                        }
                    }
                }
                if (!NoIssue)
                    IssueCount++;
            }
            for (int j = 0; j < FileCount; j++)
            {
                cout << ' ' << fixed << setw(4) << ThisFrame_Count_Blocks_Used[j];
            }
            cout << " used blocks";
            if (IssueCount)
            {
                cout << ", including " << fixed << setw(4) << IssueCount << " bad blocks";
                for (int j = 0; j < FileCount; j++)
                    Count_Frames_NotHealthy_Blocks_NotHealthy[j] += ThisFrame_Count_Blocks_NotHealthy[j];
                Blocks_NotFixed += IssueCount;
                for (int k = 0; k < FileCount; k++)
                    Count_Blocks_Used[k] += ThisFrame_Count_Blocks_Used[k];
            }
            else
            {
                Frames_Recovered++;
                Count_Blocks_Used[Priorities[0]] += 1500;
            }
        }

        fwrite(Buffers[FileCount], 120000, 1, F[FileCount]);
        cout << endl;
    }
}

void ShowFrames(size_t Count, size_t Total, const char* Next = nullptr)
{
    auto Percent = (float)Count / Total * 100;
    cout << fixed << setw(FrameCount_Width) << Count << " = " << fixed << setw(Precision + 2) << setprecision(Precision - 2) << Percent << "% of frames";
    if (Next)
        cout << Next;
}
void ShowBlocks(size_t Count, size_t Total, const char* Next = nullptr)
{
    auto Percent = (float)Count / Total * 100;
    cout << fixed << setw(BlockCount_Width) << Count << " = " << fixed << setw(Precision + 2) << setprecision(Precision - 2) << Percent << "% of blocks";
    if (Next)
        cout << Next;
}
//---------------------------------------------------------------------------
void file::DisplayEnd()
{
    if (Target.empty())
        return;

    // Check which frames are available at least once
    auto FileCount = FileNames.size();
    size_t LongestFile = 0;
    for (size_t i = 0; i < FileCount; i++)
        if (Frames_Status_Max < Frames_Status[i].size())
        {
            Frames_Status_Max = Frames_Status[i].size();
            LongestFile = i;
        }
    DisplayLines();

    cout << endl;
    FrameNumber = Frames_Status_Max;
    auto Count_Blocks_Total = Count_Blocks_Healthy[LongestFile] + Count_Blocks_NotHealthy[LongestFile];

    // Stats - Per file
    float Percent_Frames_Max = 0;
    float Percent_STA_Max = 0;
    cout << "Input files summary:" << endl;
    for (size_t i = 0; i < FileCount; i++)
    {
        auto Percent_Frames = (float)Count_Frames_Healthy[i] / FrameNumber * 100;
        auto Percent_STA = (float)Count_Blocks_Healthy[i] / Count_Blocks_Total * 100;
        if (Percent_Frames_Max < Percent_Frames)
            Percent_Frames_Max = Percent_Frames;
        if (Percent_STA_Max < Percent_STA)
            Percent_STA_Max = Percent_STA;
        cout << "File " << i << ": ";
        ShowFrames(Count_Frames_Healthy[i], FrameNumber, " are healthy; ");
        ShowBlocks(Count_Blocks_Healthy[i], Count_Blocks_Total, " are healthy.\n");
    }
    cout << endl;

    // Stats - What is used
    cout << "Usage of input files:" << endl;
    for (size_t i = 0; i < FileCount; i++)
    {
        if (Count_Blocks_Used[i])
        {
            ShowBlocks(Count_Blocks_Used[i], 1500 * FrameNumber);
            cout << " from file " << i << "     used.";
        }
        else
        {
            for (size_t j = 0; j < BlockCount_Width + Precision + 22; j++)
                cout << ' ';
            cout << "file " << i << " not used.";
        }
        cout << endl;
    }
    cout << endl;

    // Stats - Merge
    cout << "How frames are merged:" << endl;
    if (Frames_Good)
    {
        ShowFrames(Frames_Good, FrameNumber, " are recovered from full frames.\n");
    }
    if (Frames_Recovered)
    {
        ShowFrames(Frames_Recovered, FrameNumber, " are fully recovered based on reconstruction from blocks.\n");
    }
    auto Frames_Bad = FrameNumber - Frames_Good - Frames_Recovered;
    if (Frames_Bad)
    {
        size_t Count_Frames_NotHealthy_Blocks_NotHealthy_Max = -1;
        size_t Count_Frames_NotHealthy_Blocks_NotHealthy_Max_Pos;
        for (size_t i = 0; i < FileCount; i++)
        {
            if (Count_Frames_NotHealthy_Blocks_NotHealthy_Max > Count_Frames_NotHealthy_Blocks_NotHealthy[i])
            {
                Count_Frames_NotHealthy_Blocks_NotHealthy_Max = Count_Frames_NotHealthy_Blocks_NotHealthy[i];
                Count_Frames_NotHealthy_Blocks_NotHealthy_Max_Pos = i;
            }
        }
        ShowFrames(Frames_Bad, FrameNumber, " still have issues.\n                     (in these frames, from ");
        ShowBlocks(Count_Frames_NotHealthy_Blocks_NotHealthy_Max, 1500 * Frames_Bad, " bad in file ");
        cout << Count_Frames_NotHealthy_Blocks_NotHealthy_Max_Pos << "\n                                         to ";
        ShowBlocks(Blocks_NotFixed, 1500 * Frames_Bad, " bad).\n");
    }
    cout << endl;

    // Stats - Result
    auto Percent_Frames_Result = (float)(Frames_Good + Frames_Recovered) / FrameNumber * 100;
    auto Percent_STA_Result = (float)(Count_Blocks_Total - Blocks_NotFixed) / Count_Blocks_Total * 100;
    cout << "Result:" << endl;
    cout << setfill(' ') << setw(FrameCount_Width) << FrameNumber << " frames found." << endl;
    ShowFrames(Frames_Good + Frames_Recovered, FrameNumber, " are healthy; ");
    ShowBlocks(Count_Blocks_Total - Blocks_NotFixed, Count_Blocks_Total, " are healthy.\n");
    cout << "     (+" << setfill(' ') << fixed << setw(Precision + 2) << setprecision(Precision - 2) << Percent_Frames_Result - Percent_Frames_Max << ")                                 (+" << fixed << setw(Precision + 2) << setprecision(Precision - 2) << Percent_STA_Result - Percent_STA_Max << ")" << endl;
}

//---------------------------------------------------------------------------
void file::AddFrame(const MediaInfo_Event_Global_Demux_4* FrameData)
{
    if (!FrameData->StreamIDs_Size)
        return;
    auto Dseq = FrameData->StreamIDs[FrameData->StreamIDs_Size-1];
    if ((Dseq & 0xFFFF00) != ((0x2 << 16) | (0x65 << 8))) // identifier with SCT = 0x2, PackType = 0x65, and Dseq
        return;
    Dseq &= 0xFF; // Extract Dseq
    if (Dseq >= 16)
        return;
    if (FrameData->Content_Size != 4)
        return;
    
    // Store
    for (int i = 0; i < 2; i++)
    {
        // Filter unneeded content
        if (FrameData->Content[i * 2] == 0x80 && FrameData->Content[i * 2 + 1] == 0x80)
            return; // Don't store empty content
        if (Dseq && PerFrame_Captions_PerSeq_PerField.size() == 1)
        {
            auto& PerSeq0 = PerFrame_Captions_PerSeq_PerField[0];
            auto& FieldData0 = PerSeq0.FieldData[i];
            if (!FieldData0.empty() && FieldData0.back().StartFrameNumber + FieldData0.back().Captions.size() - 1 == FrameNumber && !FieldData0.back().Captions.back().compare(FrameData->Content + i * 2))
                continue; // Content is same, not stored
        }

        // Store
        if (Dseq >= PerFrame_Captions_PerSeq_PerField.size())
            PerFrame_Captions_PerSeq_PerField.resize(Dseq + 1);
        auto& PerSeq = PerFrame_Captions_PerSeq_PerField[Dseq];
        auto& FieldData = PerSeq.FieldData[i];
        if (FieldData.empty() || FieldData.back().StartFrameNumber + FieldData.back().Captions.size() != FrameNumber)
            FieldData.emplace_back(FrameNumber);
        FieldData.back().Captions.emplace_back(FrameData->Content + i * 2);
    }
}
